unit Test;

interface

uses
	Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
	Dialogs, ExtCtrls, RectUtils, Layouts;

type

	TfmTest = class;

	TMoveMode = class
	private
		FpPrevPoint: PPoint;
	public
		procedure Paint(Canvas: TCanvas; APoint: TPoint);
		procedure MoveTo(ANewPoint: TPoint);
		constructor Create;
		destructor Destroy; override;
	end;

	TfmTest = class(TForm)
		PaintBox1: TPaintBox;
    tTimer: TTimer;
		procedure FormCreate(Sender: TObject);
		procedure FormDestroy(Sender: TObject);
		procedure PaintBox1Paint(Sender: TObject);
		procedure PaintBox1MouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure tTimerTimer(Sender: TObject);
    procedure PaintBox1MouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
	private
		{ Private declarations }
		FMode: TMoveMode;
	public
		FLayout: TRTable;
		{ Public declarations }
	end;

var
  fmTest: TfmTest;

implementation

{$R *.dfm}

procedure TfmTest.FormCreate(Sender: TObject);
begin
	//DoubleBuffered := true;
	FLayout := TRTable.Create(ClientRect, 2, 2);
	FMode := nil;
end;

procedure TfmTest.FormDestroy(Sender: TObject);
begin
	if Assigned(FMode) then
		FreeAndNil(FMode);

	FLayout.Free;
end;

procedure TfmTest.PaintBox1Paint(Sender: TObject);
var i, j: integer;
		ARect: TRect;
		PaintBox: TPaintBox;
begin
	PaintBox := Sender as TPaintBox;
	FLayout.BoundsRect := ClientRect;
	PaintBox.Canvas.FillRect(ClientRect);

	for i := 0 to FLayout.ColCount - 1 do
		for j := 0 to FLayout.RowCount - 1 do
		begin
			ARect := FLayout.Cell[i,j];
			InflateRect(ARect, -1, -1);
			PaintBox.Canvas.Rectangle(ARect);
		end;
end;

{ TMoveMode }

constructor TMoveMode.Create;
begin
	inherited;
	FpPrevPoint := nil;
end;

destructor TMoveMode.Destroy;
var OldMode: TPenMode;
begin
	OldMode := fmTest.PaintBox1.Canvas.Pen.Mode;
	fmTest.PaintBox1.Canvas.Pen.Mode := pmXor;

	if Assigned(FpPrevPoint) then
	begin
		Paint(fmTest.PaintBox1.Canvas, FpPrevPoint^);
		Dispose(FpPrevPoint);
		FpPrevPoint := nil;
	end;

	fmTest.PaintBox1.Canvas.Pen.Mode := OldMode;

	inherited;
end;

procedure TMoveMode.MoveTo(ANewPoint: TPoint);
var OldMode: TPenMode;
begin
	OldMode := fmTest.PaintBox1.Canvas.Pen.Mode;
	fmTest.PaintBox1.Canvas.Pen.Mode := pmXor;

	if Assigned(FpPrevPoint) then
	begin
		Paint(fmTest.PaintBox1.Canvas, FpPrevPoint^);
		FpPrevPoint^ := ANewPoint;
	end;

	Paint(fmTest.PaintBox1.Canvas, ANewPoint);

	if not Assigned(FpPrevPoint) then
	begin
		New(FpPrevPoint);
		FpPrevPoint^ := ANewPoint;
	end;

	fmTest.PaintBox1.Canvas.Pen.Mode := OldMode;
end;

procedure TfmTest.PaintBox1MouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
begin
	if not Assigned(FMode) then
	begin
		FMode := TMoveMode.Create;
	end
	else
		FMode.MoveTo(Point(X,Y));
end;

procedure TfmTest.tTimerTimer(Sender: TObject);
var Pt, CliPt: TPoint;
		hWindow: HWND;
begin
	inherited;

	Pt := Mouse.CursorPos;
	hWindow := WindowFromPoint(Pt);

	if (hWindow <> Self.Handle) then
	begin
		if Assigned(FMode) then
			FreeAndNil(FMode);
	end
	else
	begin
		CliPt := ScreenToClient(Pt);
		If not PtInRect(PaintBox1.BoundsRect, CliPt) then
		begin
			if Assigned(FMode) then
				FreeAndNil(FMode);
		end;
	end;
end;

procedure TMoveMode.Paint(Canvas: TCanvas; APoint: TPoint);
var ARect: TRect;
    MappedRect: TRect;
begin
	ARect := Rect(0,0,100,100);
	Center(ARect, APoint);

  MappedRect := MapRc2Rc(ARect, fmTest.FLayout.BoundsRect, fmTest.FLayout.Cell[0,1]);

	Canvas.DrawFocusRect(MappedRect);
	Canvas.DrawFocusRect(ARect);
end;

procedure TfmTest.PaintBox1MouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
	if Assigned(FMode) then
	begin
		FreeAndNil(FMode);
	end;
end;

end.
