(******************************************************************************
	Some routines to easy manipulate with
	rectangles, points and ranges.
  Written by Itby. 2006-2009.
 ******************************************************************************)
unit RectUtils;

interface

uses Windows, Types, DeCAL, Classes, Graphics, Controls;

type
	// A point with real coords.
	TFloatPoint = packed record
		X: double;
		Y: double;
	end;

	PFloatRect = ^TFloatRect;
	TFloatRect = packed record
		case integer of
			0: (Left, Top, Right, Bottom: double);
			1: (TopLeft, BottomRight: TFloatPoint);
	end;

	// Range manipulation data structures
	TIntRange = packed record
		MinVal: Integer;
		MaxVal: Integer;
	end;

	PFloatRange = ^TFloatRange;
	TFloatRange = packed record
		MinVal: double;
		MaxVal: double;
	end;

  ////////////////////////////////////////////////////////////////////////
	//
	// Helper functions to simplify manipulation with rectangles, sizes
	// and points.
	//
	////////////////////////////////////////////////////////////////////////

	function FloatPoint(aX, aY: double): TFloatPoint;

	// Convert the point to key and back (for hashing functions)
	function Pt2Key(Pt: TPoint): Integer;
	function Key2Pt(Key: Integer): TPoint;

	// Get width and height of the rect correspondingly.
	function RHeight(rcRect: TRect): Integer; overload;
	function RWidth (rcRect: TRect): Integer; overload;

	function RHeight(rcRect: TFloatRect): double; overload;
	function RWidth (rcRect: TFloatRect): double; overload;

	// Set height and width for the rectangles.
	procedure SetRHeight(var rc: TRect; Height: Integer);
	procedure SetRWidth(var rc: TRect; Width: Integer);

	// Move the rect to the point.
	procedure MoveRcToPt(var rcRect: TRect; Pt: TPoint);

	// Construct the size object.
	function Size(cx, cy: Integer): TSize;
	// Get size of the rect.
	function RectSize(rc: TRect): TSize;
	// Construct the rect with the size.
	function RectWithSize(sz: TSize): TRect; overload;
  function RectWithSize(cx, cy: Integer): TRect; overload;
	function RectWithSize(sz: TSize; BasePt: TPoint): TRect; overload;

  function Pt2Rect(Pt: TPoint): TRect;

	function CreateRect(ATopLeft: TPoint; ABottomRight: TPoint): TRect; overload;

	function FloatRect(ATopLeft: TFloatPoint; ABottomRight: TFloatPoint): TFloatRect; overload;
	function FloatRect(ALeft, ATop, ARight, ABottom: double): TFloatRect; overload;
	function FloatRect(AHRange, AVRange: TFloatRange): TFloatRect; overload;

	procedure InvertRectV(var rc: TRect);
	procedure InvertRectH(var rc: TRect);

	// Convert size to the corresponding point (relative to (0,0) pos, of course).
	function SizeToPoint(sz: TSize): TPoint;
	function PointToSize(pt: TPoint): TSize;

	procedure RPressToTop(var rcRect: TRect; Top: Integer);
	procedure RPressToBottom(var rcRect: TRect; Bottom: Integer);
	procedure RPressToLeft(var rcRect: TRect; Left: Integer);
	procedure RPressToRight(var rcRect: TRect; Right: Integer);

	function NormalizeRect(rc: TRect): TRect;
	function CreateNormRect(pt1, pt2: TPoint): TRect;

	// Shift the point
	function PtShift(Pt: TPoint; Sz: TSize): TPoint;

	// Distance between two points
	function PtDistance(Pt1: TPoint; Pt2: TPoint): integer; overload;
	function PtDistance(Pt1: TFloatPoint; Pt2: TFloatPoint): double; overload;

	// Divide the size into equal parts.
	function DivSize(sz: TSize; Divs: TPoint): TSize;

	// Define if the sizes are equal.
	function SzEq(sz1, sz2: TSize): boolean;
	function REq(rc1, rc2: TRect): boolean;
	function PtEq(pt1, pt2: TPoint): boolean;

	// Center the rect relative to other rect.
	procedure Center(var RectToBeCentered: TRect; const CentralRect: TRect); overload;
  procedure Center(var RectToBeCentered: TRect; const CentralPt: TPoint); overload;

	procedure HCenter(var RectToBeCentered: TRect; const CentralRect: TRect);
	procedure VCenter(var RectToBeCentered: TRect; const CentralRect: TRect);

	// Get vertical center line (returned as rect)
	function RVCenter(const rcRect: TRect): TRect;
	// Get horizontal center line (returned as rect)
	function RHCenter(const rcRect: TRect): TRect;

	// Range manipulation routines.
	function IntRange(AMinVal, AMaxVal: Integer): TIntRange;
	function FloatRange(AMinVal, AMaxVal: double): TFloatRange;
	function AsFloatRange(AIntRange: TIntRange): TFloatRange;

	function RVRange(rcRect: TRect): TIntRange; overload;
	function RHRange(rcRect: TRect): TIntRange; overload;

	function RVRange(rcRect: TFloatRect): TFloatRange; overload;
	function RHRange(rcRect: TFloatRect): TFloatRange; overload;

	function RangeValue(ARange: TIntRange): Integer; overload;
	function RangeValue(ARange: TFloatRange): double; overload;

	function Normalize(Value: Integer; ARange: TIntRange): double; overload;
	function Normalize(Value: double; ARange: TFloatRange): double; overload;

	function MapVal2Range(Value: Integer; FromRange: TIntRange; ToRange: TFloatRange): double; overload;
	function MapVal2Range(Value: Integer; FromRange: TIntRange; ToRange: TIntRange): integer; overload;
	function MapVal2Range(Value: double; FromRange: TFloatRange; ToRange: TFloatRange): double; overload;
	function MapVal2Range(Value: double; FromRange: TFloatRange; ToRange: TIntRange): Integer; overload;

	function MapDelta2Range(Value: Integer; FromRange: TIntRange; ToRange: TFloatRange): double; overload;
	function MapDelta2Range(Value: Integer; FromRange: TIntRange; ToRange: TIntRange): integer; overload;
	function MapDelta2Range(Value: double; FromRange: TFloatRange; ToRange: TFloatRange): double; overload;
	function MapDelta2Range(Value: double; FromRange: TFloatRange; ToRange: TIntRange): Integer; overload;

	function MapPoint(const Pt: TFloatPoint; const SrcRect: TFloatRect; const RectToMap: TRect): TPoint; overload;
	function MapPoint(const Pt: TPoint; const SrcRect: TRect; const RectToMap: TFloatRect): TFloatPoint; overload;
	function MapPoint(const Pt: TPoint; const SrcRect: TRect; const RectToMap: TRect): TPoint; overload;

	function MapRc2Rc(const RcToBeMapped: TFloatRect; const RcMapFrom: TFloatRect; const RcMapTo: TRect): TRect; overload;
	function MapRc2Rc(const RcToBeMapped: TRect; const RcMapFrom: TRect; const RcMapTo: TFloatRect): TFloatRect; overload;
	function MapRc2Rc(const RcToBeMapped: TRect; const RcMapFrom: TRect; const RcMapTo: TRect): TRect; overload;

	function FloatPt2IntPt(FloatPt: TFloatPoint): TPoint;
	function FloatRc2IntRc(FloatRect: TFloatRect): TRect;

	function ClipVal2Range(Value: double; AClipRange: TFloatRange): double;

	function IntersectRange(Range1: TIntRange; Range2: TIntRange): TIntRange; overload;
	function IntersectRange(Range1: TIntRange; Range2: TFloatRange): TFloatRange; overload;
	function IntersectRange(Range1: TFloatRange; Range2: TFloatRange): TFloatRange; overload;

	// Abs for values having double precision.
	function AbsD(const Val: double): double;

implementation

uses SysUtils, Math;

function FloatPoint(aX, aY: double): TFloatPoint;
begin
	Result.X := aX;
	Result.Y := aY;
end;

function RHeight(rcRect: TRect): Integer;
begin
	Result := Abs(rcRect.Bottom - rcRect.Top);
end;

function RWidth (rcRect: TRect): Integer;
begin
	Result := Abs(rcRect.Right - rcRect.Left);
end;

function RHeight(rcRect: TFloatRect): double;
begin
	Result := Abs(rcRect.Bottom - rcRect.Top);
end;

function RWidth (rcRect: TFloatRect): double;
begin
	Result := Abs(rcRect.Right - rcRect.Left);
end;

procedure SetRHeight(var rc: TRect; Height: Integer);
begin
	rc.BottomRight := PtShift(rc.TopLeft, Size(RWidth(rc), Height));
end;

procedure SetRWidth(var rc: TRect; Width: Integer);
begin
	rc.BottomRight := PtShift(rc.TopLeft, Size(Width, RHeight(rc)));
end;

function Size(cx, cy: Integer): TSize;
begin
  Result.cx := cx;
	Result.cy := cy;
end;

function RectSize(rc: TRect): TSize;
begin
	Result := Size(RWidth(rc), RHeight(rc));
end;

function RectWithSize(sz: TSize): TRect;
begin
	FillMemory(@Result, sizeof(Result), 0);
	Result.BottomRight := SizeToPoint(sz);
end;

function RectWithSize(cx, cy: Integer): TRect;
begin
	FillMemory(@Result, sizeof(Result), 0);
	Result.BottomRight := Point(cx,cy);
end;

function RectWithSize(sz: TSize; BasePt: TPoint): TRect;
var TheRect: TRect;
begin
	TheRect := RectWithSize(sz);
	MoveRcToPt(TheRect, BasePt);
end;

function Pt2Rect(Pt: TPoint): TRect;
begin
  Result := CreateRect(Pt,Pt);
end;

function CreateRect(ATopLeft: TPoint; ABottomRight: TPoint): TRect; overload;
begin
	Result.TopLeft := ATopLeft;
	Result.BottomRight := ABottomRight;
end;

function FloatRect(ATopLeft: TFloatPoint; ABottomRight: TFloatPoint): TFloatRect; overload;
begin
	Result.TopLeft := ATopLeft;
	Result.BottomRight := ABottomRight;
end;

function FloatRect(ALeft, ATop, ARight, ABottom: double): TFloatRect; overload;
begin
	Result.Left := ALeft;
	Result.Right := ARight;
	Result.Top := ATop;
	Result.Bottom := ABottom;
end;

function FloatRect(AHRange, AVRange: TFloatRange): TFloatRect;
begin
	Result.Left := AHRange.MinVal;
	Result.Right := AHRange.MaxVal;
	Result.Top := AVRange.MinVal;
	Result.Bottom := AVRange.MaxVal;
end;

procedure InvertRectV(var rc: TRect);
var Temp: Integer;
begin
	Temp := rc.Top;
	rc.Top := rc.Bottom;
	rc.Bottom := Temp;
end;

procedure InvertRectH(var rc: TRect);
var Temp: Integer;
begin
	Temp := rc.Left;
	rc.Left := rc.Right;
	rc.Right := Temp;
end;

function SizeToPoint(sz: TSize): TPoint;
begin
  Result := TPoint(sz);
end;

function PointToSize(pt: TPoint): TSize;
begin
  Result := TSize(pt);
end;

procedure MoveRcToPt(var rcRect: TRect; Pt: TPoint);
var Width, Height: Integer;
begin
  Width := RWidth(rcRect);
  Height := RHeight(rcRect);
  rcRect.Left := Pt.X;
  rcRect.Top := Pt.Y;

  rcRect.Right := rcRect.Left + Width;
  rcRect.Bottom := rcRect.Top + Height;
end;

procedure RPressToTop(var rcRect: TRect; Top: Integer);
var Height: Integer;
begin
  Height := RHeight(rcRect);

  rcRect.Top := Top;
  rcRect.Bottom := Top + Height;
end;

procedure RPressToBottom(var rcRect: TRect; Bottom: Integer);
var Height: Integer;
begin
	Height := RHeight(rcRect);

	rcRect.Bottom := Bottom;
	rcRect.Top := Bottom - Height;
end;

procedure RPressToLeft(var rcRect: TRect; Left: Integer);
var Width: Integer;
begin
	Width := RWidth(rcRect);

	rcRect.Left := Left;
	rcRect.Right := Left + Width;
end;

procedure RPressToRight(var rcRect: TRect; Right: Integer);
var Width: Integer;
begin
	Width := RWidth(rcRect);

	rcRect.Right := Right;
	rcRect.Left := Right - Width;
end;

function NormalizeRect(rc: TRect): TRect;
begin
	Result.Left := rc.Left;
	Result.Right := rc.Right;

	if rc.Right < rc.Left then
	begin
		Result.Left := rc.Right;
		Result.Right := rc.Left;
	end;

	Result.Bottom := rc.Bottom;
	Result.Top := rc.Top;

	if rc.Bottom < rc.Top then
	begin
		Result.Bottom := rc.Top;
		Result.Top := rc.Bottom;
	end;
end;

function CreateNormRect(pt1, pt2: TPoint): TRect;
begin
	Result := Rect(pt1, pt2);
	Result := NormalizeRect(Result);
end;

function PtShift(Pt: TPoint; Sz: TSize): TPoint;
begin
	Result.X := Pt.X + Sz.cx;
	Result.Y := Pt.Y + Sz.cy;
end;

function PtDistance(Pt1: TPoint; Pt2: TPoint): integer;
var R: TRect;
begin
	R := Rect(Pt1, Pt2);
	Result := round(sqrt(RHeight(R)*RHeight(R) + RWidth(R)*RWidth(R)));
end;

function PtDistance(Pt1: TFloatPoint; Pt2: TFloatPoint): double;
var R: TFloatRect;
begin
	R := FloatRect(Pt1, Pt2);
	Result := sqrt(RHeight(R)*RHeight(R) + RWidth(R)*RWidth(R));
end;

function DivSize(sz: TSize; Divs: TPoint): TSize;
begin
	Result := Size(sz.cx div Divs.X, sz.cy div Divs.Y);
end;

function SzEq(sz1, sz2: TSize): boolean;
begin
	Result := (sz1.cx = sz2.cx) and (sz1.cy = sz2.cy);
end;

function REq(rc1, rc2: TRect): boolean;
begin
	Result := (rc1.Left = rc2.Left) and (rc1.Top = rc2.Top) and (rc1.Right = rc2.Right) and (rc1.Bottom = rc2.Bottom);
end;

function PtEq(pt1, pt2: TPoint): boolean;
begin
	Result := (pt1.X = pt2.X) and (pt1.Y = pt2.Y);
end;

function RHCenter(const rcRect: TRect): TRect;
begin
	Result.Top := rcRect.Top;
	Result.Bottom := rcRect.Bottom;

	Result.Left := (RWidth(rcRect) shr 1) + rcRect.Left;
	Result.Right := Result.Left;
end;

function RVCenter(const rcRect: TRect): TRect;
begin
	Result.Left := rcRect.Left;
	Result.Right := rcRect.Right;

	Result.Top := (RHeight(rcRect) shr 1) + rcRect.Top;
	Result.Bottom := Result.Top;
end;

procedure Center(var RectToBeCentered: TRect; const CentralRect: TRect);
begin
	HCenter(RectToBeCentered, CentralRect);
	VCenter(RectToBeCentered, CentralRect);
end;

procedure Center(var RectToBeCentered: TRect; const CentralPt: TPoint);
begin
	HCenter(RectToBeCentered, Pt2Rect(CentralPt));
	VCenter(RectToBeCentered, Pt2Rect(CentralPt));
end;

procedure HCenter(var RectToBeCentered: TRect; const CentralRect: TRect);
var HCenterLine: TRect;
		RectTBCWidth: Integer;
begin
	HCenterLine := RHCenter(CentralRect);

	RectTBCWidth := RWidth(RectToBeCentered);

	RectToBeCentered.Left   := HCenterLine.Left - (RectTBCWidth shr 1);
	RectToBeCentered.Right  := HCenterLine.Right + (RectTBCWidth shr 1);
end;

procedure VCenter(var RectToBeCentered: TRect; const CentralRect: TRect);
var VCenterLine: TRect;
		RectTBCHeight: Integer;
begin
	VCenterLine := RVCenter(CentralRect);

	RectTBCHeight := RHeight(RectToBeCentered);

	RectToBeCentered.Top    := VCenterLine.Top - (RectTBCHeight shr 1);
	RectToBeCentered.Bottom := VCenterLine.Bottom + (RectTBCHeight shr 1);
end;

function Key2Pt(Key: Integer): TPoint;
begin
	Result.X := Smallint(Key shr (sizeof(Smallint)*8));
	Result.Y := Smallint(Key);
end;

function Pt2Key(Pt: TPoint): Integer;
begin
	Result := (Smallint(Pt.X) shl (sizeof(Smallint)*8)) or (Smallint(Pt.Y));
end;

function IntRange(AMinVal, AMaxVal: Integer): TIntRange;
begin
	Result.MinVal := AMinVal;
	Result.MaxVal := AMaxVal;
end;

function FloatRange(AMinVal, AMaxVal: double): TFloatRange;
begin
	Result.MinVal := AMinVal;
	Result.MaxVal := AMaxVal;
end;

function AsFloatRange(AIntRange: TIntRange): TFloatRange;
begin
	Result.MinVal := AIntRange.MinVal;
	Result.MaxVal := AIntRange.MaxVal;
end;

function RVRange(rcRect: TRect): TIntRange;
begin
	Result := IntRange(rcRect.Top, rcRect.Bottom);
end;

function RHRange(rcRect: TRect): TIntRange;
begin
	Result := IntRange(rcRect.Left, rcRect.Right);
end;

function RVRange(rcRect: TFloatRect): TFloatRange; overload;
begin
	Result := FloatRange(rcRect.Top, rcRect.Bottom);
end;

function RHRange(rcRect: TFloatRect): TFloatRange; overload;
begin
	Result := FloatRange(rcRect.Left, rcRect.Right);
end;

function RangeValue(ARange: TIntRange): Integer;
begin
	Result := ARange.MaxVal - ARange.MinVal;
end;

function RangeValue(ARange: TFloatRange): double;
begin
	Result := ARange.MaxVal - ARange.MinVal;
end;

function Normalize(Value: Integer; ARange: TIntRange): double;
begin
	Result := (Value - ARange.MinVal)/RangeValue(ARange);
end;

function Normalize(Value: double; ARange: TFloatRange): double;
begin
	Result := (Value - ARange.MinVal)/RangeValue(ARange);
end;

function MapVal2Range(Value: Integer; FromRange: TIntRange; ToRange: TFloatRange): double;
begin
	Result := Normalize(Value, FromRange)*RangeValue(ToRange)+ToRange.MinVal;
end;

function MapVal2Range(Value: Integer; FromRange: TIntRange; ToRange: TIntRange): integer; overload;
begin
	Result := Round(Normalize(Value, FromRange)*RangeValue(ToRange)+ToRange.MinVal);
end;

function MapVal2Range(Value: double; FromRange: TFloatRange; ToRange: TFloatRange): double;
begin
	Result := Normalize(Value, FromRange)*RangeValue(ToRange)+ToRange.MinVal;
end;

function MapVal2Range(Value: double; FromRange: TFloatRange; ToRange: TIntRange): Integer;
begin
	Result := Round(Normalize(Value, FromRange)*RangeValue(ToRange)+ToRange.MinVal);
end;

function MapDelta2Range(Value: Integer; FromRange: TIntRange; ToRange: TFloatRange): double; overload;
begin
	Result := Value*RangeValue(ToRange)/RangeValue(FromRange);
end;

function MapDelta2Range(Value: Integer; FromRange: TIntRange; ToRange: TIntRange): integer; overload;
begin
	Result := Round(Value*RangeValue(ToRange)/RangeValue(FromRange));
end;

function MapDelta2Range(Value: double; FromRange: TFloatRange; ToRange: TFloatRange): double; overload;
begin
	Result := Value*RangeValue(ToRange)/RangeValue(FromRange);
end;

function MapDelta2Range(Value: double; FromRange: TFloatRange; ToRange: TIntRange): Integer; overload;
begin
	Result := Round(Value*RangeValue(ToRange)/RangeValue(FromRange));
end;

function MapPoint(const Pt: TFloatPoint; const SrcRect: TFloatRect; const RectToMap: TRect): TPoint;
begin
	Result.X := MapVal2Range(Pt.X, RHRange(SrcRect), RHRange(RectToMap));
	Result.Y := MapVal2Range(Pt.Y, RVRange(SrcRect), RVRange(RectToMap));
end;

function MapPoint(const Pt: TPoint; const SrcRect: TRect; const RectToMap: TFloatRect): TFloatPoint;
begin
	Result.X := MapVal2Range(Pt.X, RHRange(SrcRect), RHRange(RectToMap));
	Result.Y := MapVal2Range(Pt.Y, RVRange(SrcRect), RVRange(RectToMap));
end;

function MapPoint(const Pt: TPoint; const SrcRect: TRect; const RectToMap: TRect): TPoint;
begin
  Result.X := MapVal2Range(Pt.X, RHRange(SrcRect), RHRange(RectToMap));
  Result.Y := MapVal2Range(Pt.Y, RVRange(SrcRect), RVRange(RectToMap));
end;

function MapRc2Rc(const RcToBeMapped: TRect; const RcMapFrom: TRect; const RcMapTo: TFloatRect): TFloatRect;
begin
	Result.TopLeft := MapPoint(RcToBeMapped.TopLeft, RcMapFrom, RcMapTo);
	Result.BottomRight := MapPoint(RcToBeMapped.BottomRight, RcMapFrom, RcMapTo);
end;

function MapRc2Rc(const RcToBeMapped: TFloatRect; const RcMapFrom: TFloatRect; const RcMapTo: TRect): TRect;
begin
	Result.TopLeft := MapPoint(RcToBeMapped.TopLeft, RcMapFrom, RcMapTo);
	Result.BottomRight := MapPoint(RcToBeMapped.BottomRight, RcMapFrom, RcMapTo);
end;

function MapRc2Rc(const RcToBeMapped: TRect; const RcMapFrom: TRect; const RcMapTo: TRect): TRect;
begin
  Result.TopLeft := MapPoint(RcToBeMapped.TopLeft, RcMapFrom, RcMapTo);
	Result.BottomRight := MapPoint(RcToBeMapped.BottomRight, RcMapFrom, RcMapTo);
end;

function FloatPt2IntPt(FloatPt: TFloatPoint): TPoint;
begin
  Result.X := Round(FloatPt.X);
  Result.Y := Round(FloatPt.Y);
end;

function FloatRc2IntRc(FloatRect: TFloatRect): TRect;
begin
  Result.TopLeft := FloatPt2IntPt(FloatRect.TopLeft);
  Result.BottomRight := FloatPt2IntPt(FloatRect.BottomRight);
end;

function ClipVal2Range(Value: double; AClipRange: TFloatRange): double;
begin
	Result := EnsureRange(Value, AClipRange.MinVal, AClipRange.MaxVal);
end;

function IntersectRange(Range1: TIntRange; Range2: TIntRange): TIntRange;
begin
	Result.MinVal := Max(Range1.MinVal, Range2.MinVal);
	Result.MaxVal := Min(Range1.MaxVal, Range2.MaxVal);
end;

function IntersectRange(Range1: TIntRange; Range2: TFloatRange): TFloatRange; overload;
begin
	Result.MinVal := Max(Range1.MinVal, Range2.MinVal);
	Result.MaxVal := Min(Range1.MaxVal, Range2.MaxVal);
end;

function IntersectRange(Range1: TFloatRange; Range2: TFloatRange): TFloatRange; overload;
begin
	Result.MinVal := Max(Range1.MinVal, Range1.MaxVal);
	Result.MaxVal := Min(Range1.MinVal, Range1.MaxVal);
end;

function AbsD(const Val: double): double;
begin
	Result := Val;
	if Val < 0 then
		Result := -Result;
end;

end.
